/*
 * Comment To ID
 * Copyright (C) 2005 Kevin Kofler
 *
 * Scale3x effect from the Scale2x project
 * Copyright (C) 2001, 2002, 2003, 2004 Andrea Mazzoleni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "dataext.h"
#include <tigcclib.h>

typedef unsigned char scale3x_uint8;

static inline void scale3x_8_def_whole(scale3x_uint8* restrict dst0, scale3x_uint8* restrict dst1, scale3x_uint8* restrict dst2, const scale3x_uint8* restrict src0, const scale3x_uint8* restrict src1, const scale3x_uint8* restrict src2, unsigned count)
{
	/* first pixel */
	if (src0[0] != src2[0] && src1[0] != src1[1]) {
		dst0[0] = src1[0];
		dst0[1] = (src1[0] == src0[0] && src1[0] != src0[1]) || (src1[1] == src0[0] && src1[0] != src0[0]) ? src0[0] : src1[0];
		dst0[2] = src1[1] == src0[0] ? src1[1] : src1[0];
		dst1[0] = (src1[0] == src0[0] && src1[0] != src2[0]) || (src1[0] == src2[0] && src1[0] != src0[0]) ? src1[0] : src1[0];
		dst1[1] = src1[0];
		dst1[2] = (src1[1] == src0[0] && src1[0] != src2[1]) || (src1[1] == src2[0] && src1[0] != src0[1]) ? src1[1] : src1[0];
		dst2[0] = src1[0];
		dst2[1] = (src1[0] == src2[0] && src1[0] != src2[1]) || (src1[1] == src2[0] && src1[0] != src2[0]) ? src2[0] : src1[0];
		dst2[2] = src1[1] == src2[0] ? src1[1] : src1[0];
	} else {
		dst0[0] = src1[0];
		dst0[1] = src1[0];
		dst0[2] = src1[0];
		dst1[0] = src1[0];
		dst1[1] = src1[0];
		dst1[2] = src1[0];
		dst2[0] = src1[0];
		dst2[1] = src1[0];
		dst2[2] = src1[0];
	}
	++src0;
	++src1;
	++src2;
	dst0 += 3;
	dst1 += 3;
	dst2 += 3;

	/* central pixels */
	count -= 2;
	while (count) {
		if (src0[0] != src2[0] && src1[-1] != src1[1]) {
			dst0[0] = src1[-1] == src0[0] ? src1[-1] : src1[0];
			dst0[1] = (src1[-1] == src0[0] && src1[0] != src0[1]) || (src1[1] == src0[0] && src1[0] != src0[-1]) ? src0[0] : src1[0];
			dst0[2] = src1[1] == src0[0] ? src1[1] : src1[0];
			dst1[0] = (src1[-1] == src0[0] && src1[0] != src2[-1]) || (src1[-1] == src2[0] && src1[0] != src0[-1]) ? src1[-1] : src1[0];
			dst1[1] = src1[0];
			dst1[2] = (src1[1] == src0[0] && src1[0] != src2[1]) || (src1[1] == src2[0] && src1[0] != src0[1]) ? src1[1] : src1[0];
			dst2[0] = src1[-1] == src2[0] ? src1[-1] : src1[0];
			dst2[1] = (src1[-1] == src2[0] && src1[0] != src2[1]) || (src1[1] == src2[0] && src1[0] != src2[-1]) ? src2[0] : src1[0];
			dst2[2] = src1[1] == src2[0] ? src1[1] : src1[0];
		} else {
			dst0[0] = src1[0];
			dst0[1] = src1[0];
			dst0[2] = src1[0];
			dst1[0] = src1[0];
			dst1[1] = src1[0];
			dst1[2] = src1[0];
			dst2[0] = src1[0];
			dst2[1] = src1[0];
			dst2[2] = src1[0];
		}

		++src0;
		++src1;
		++src2;
		dst0 += 3;
		dst1 += 3;
		dst2 += 3;
		--count;
	}

	/* last pixel */
	if (src0[0] != src2[0] && src1[-1] != src1[0]) {
		dst0[0] = src1[-1] == src0[0] ? src1[-1] : src1[0];
		dst0[1] = (src1[-1] == src0[0] && src1[0] != src0[0]) || (src1[0] == src0[0] && src1[0] != src0[-1]) ? src0[0] : src1[0];
		dst0[2] = src1[0];
		dst1[0] = (src1[-1] == src0[0] && src1[0] != src2[-1]) || (src1[-1] == src2[0] && src1[0] != src0[-1]) ? src1[-1] : src1[0];
		dst1[1] = src1[0];
		dst1[2] = (src1[0] == src0[0] && src1[0] != src2[0]) || (src1[0] == src2[0] && src1[0] != src0[0]) ? src1[0] : src1[0];
		dst2[0] = src1[-1] == src2[0] ? src1[-1] : src1[0];
		dst2[1] = (src1[-1] == src2[0] && src1[0] != src2[0]) || (src1[0] == src2[0] && src1[0] != src2[-1]) ? src2[0] : src1[0];
		dst2[2] = src1[0];
	} else {
		dst0[0] = src1[0];
		dst0[1] = src1[0];
		dst0[2] = src1[0];
		dst1[0] = src1[0];
		dst1[1] = src1[0];
		dst1[2] = src1[0];
		dst2[0] = src1[0];
		dst2[1] = src1[0];
		dst2[2] = src1[0];
	}
}

#define SSDST(bits, num) (scale3x_uint##bits *)dst##num
#define SSSRC(bits, num) (const scale3x_uint##bits *)src##num

static inline void stage_scale3x(void* dst0, void* dst1, void* dst2, const void* src0, const void* src1, const void* src2, unsigned pixel_per_row)
{
	scale3x_8_def_whole(SSDST(8,0), SSDST(8,1), SSDST(8,2), SSSRC(8,0), SSSRC(8,1), SSSRC(8,2), pixel_per_row);
}

#define SCDST(i) (dst+(i)*dst_slice)
#define SCSRC(i) (src+(i)*src_slice)

static void scale3x(void* void_dst, unsigned dst_slice, const void* void_src, unsigned src_slice, unsigned width, unsigned height)
{
	unsigned char* dst = (unsigned char*)void_dst;
	const unsigned char* src = (const unsigned char*)void_src;
	unsigned count;

	count = height;

	stage_scale3x(SCDST(0), SCDST(1), SCDST(2), SCSRC(0), SCSRC(0), SCSRC(1), width);

	dst = SCDST(3);

	count -= 2;
	while (count) {
		stage_scale3x(SCDST(0), SCDST(1), SCDST(2), SCSRC(0), SCSRC(1), SCSRC(2), width);

		dst = SCDST(3);
		src = SCSRC(1);

		--count;
	}

	stage_scale3x(SCDST(0), SCDST(1), SCDST(2), SCSRC(0), SCSRC(1), SCSRC(1), width);
}

static void scale3x_ICON(const ICON *src, unsigned char dest[48][48])
{
 unsigned char srcx[16][16];
 for (unsigned i=0;i<16;i++)
  for (unsigned j=0;j<16;j++)
   srcx[i][j]=peek_bit((unsigned char*)src+(signed)(i+i+(j>>3)),(7-j)&7);
 scale3x(dest,48,srcx,16,16,16);
}

static unsigned char croptbl[22]={1,4,6,8,10,12,14,16,18,20,22,
                                  25,27,29,31,33,35,37,39,41,43,46};

static void crop_48x48_22x22(unsigned char src[48][48],BITMAP *dest)
{
 dest->NumRows = dest->NumCols = 22;
 memset(dest->Data,0,66);
 for (unsigned i=0;i<22;i++)
  for (unsigned j=0;j<22;j++)
   if (src[croptbl[i]][croptbl[j]])
    poke_bset(dest->Data+(signed)(i*3+(j>>3)),7-j);
}

void fwrite_throw(const void *ptr, unsigned short n, FILE *stream)
{
 if (fwrite(ptr,1,n,stream)<n) ER_throw(ER_MEMORY);
}

void _main(void)
{
 FILE *f=fopen("idicocnv","wb");
 if (!f) ER_throw(ER_PROTECTED);
 TRY
  unsigned short c=0;
  fwrite_throw((unsigned short []){0x7069u,0x2F38u,0u},6,f);
  for (SYM_ENTRY *symentry=SymFindFirst(NULL,FO_RECURSE|FO_SKIP_TEMPS);symentry;
       symentry=SymFindNext()) {
   const char *filename=symentry->name;
   HANDLE h=symentry->handle;
   if (!h) continue;
   void *fptr=HeapDeref(h);
   if (fptr && *HToESI(h)==ASM_TAG) {
    fptr+=2;
    DATA_EXT *deptr=DataExtensionPointer(fptr);
    if (deptr) {
     ICON *iconptr=SearchDataExtension(deptr,DE_BW_ICON);
     if (!iconptr) iconptr=SearchDataExtension(deptr,DE_GRAYSCALE_ICON);
     if (iconptr) {
      unsigned char buf[48][48];
      unsigned char bmp[70];
      scale3x_ICON(iconptr,buf);
      crop_48x48_22x22(buf,(BITMAP *)bmp);
      const char *progname=GetDEProgramNameEx(deptr,filename);
      size_t l=(strlen(progname)+2ul)&(~1ul);
      unsigned char identry[80+l];
      strncpy(identry,filename,8);
      *(unsigned short*)(identry+8)=70+l;
      memcpy(identry+10,bmp,70);
      strncpy(identry+80,progname,l);
      fwrite_throw(identry,80+l,f);
      c++;
     }
    }
   }
  }
  fwrite_throw((ESQ[]){0,'i','c','o','n',0,OTH_TAG},7,f);
  fseek(f,4,SEEK_SET);
  fwrite_throw(&c,2,f);
 FINALLY
  fclose(f);
 ENDFINAL
}
